from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.widgets import Button, Label

from ..app_config import app_config


class Check(Label):
    def __init__(self) -> None:
        variation = app_config.variation

        self.default_text = (
            f"{app_config.ui['check_default_text']} " * variation["num_pegs"]
        )[:-1]
        self.hover_text = (
            f"{app_config.ui['check_hover_text']} " * variation["num_pegs"]
        )[:-1]

        super().__init__(self.default_text, id="check", classes="check")

    def on_enter(self) -> None:
        self.update(self.hover_text)

    def on_leave(self) -> None:
        self.update(self.default_text)


class Row(Horizontal):
    def __init__(self, row_number: int) -> None:
        super().__init__(classes="row")

        self.row_number = row_number

        variation = app_config.variation

        self.code_pegs: list[Button] = [
            Button(label=app_config.ui["code_blank_symbol"], classes="code_peg")
            for _ in range(variation["num_pegs"])
        ]

        self.check: Check = Check()

    def compose(self) -> ComposeResult:
        yield Label(f"{self.row_number:02}", classes="num")
        for code_peg in self.code_pegs:
            yield code_peg
        yield self.check


class Board(VerticalScroll):
    def __init__(self) -> None:
        super().__init__()

        self.current_row_number = 1
        self.current_row = Row(row_number=self.current_row_number)

    def compose(self) -> ComposeResult:
        yield self.current_row

    def add_row(self):
        self.current_row_number += 1
        self.current_row = Row(row_number=self.current_row_number)
        self.mount(self.current_row)
        self.scroll_end()



import random

from .app_config import app_config


class Game:
    def __init__(self) -> None:
        variation = app_config.variation

        self.num_rows = variation["num_rows"]
        self.num_pegs = variation["num_pegs"]
        self.num_symbols = variation["num_symbols"]

        symbols: list[int] = list(range(1, self.num_symbols + 1))
        if app_config.settings.blank_symbol:
            symbols.append(0)

        self.maker_code: list[int]
        if app_config.settings.duplicate_symbols:
            self.maker_code = random.choices(symbols, k=self.num_pegs)
        else:
            self.maker_code = random.sample(symbols, k=self.num_pegs)

    def check_code(self, breaker_code: list[int]) -> tuple[int, int]:
        breaker_code_no_reds = [
            symbol
            for i, symbol in enumerate(breaker_code)
            if self.maker_code[i] != breaker_code[i]
        ]
        maker_code_no_reds = [
            symbol
            for i, symbol in enumerate(self.maker_code)
            if self.maker_code[i] != breaker_code[i]
        ]

        num_red_pegs: int = len(breaker_code) - len(breaker_code_no_reds)

        num_white_pegs: int = 0
        for symbol in breaker_code_no_reds:
            if symbol in maker_code_no_reds:
                num_white_pegs += 1
                maker_code_no_reds.remove(symbol)

        return num_red_pegs, num_white_pegs

    def get_maker_code(self) -> list[int]:
        return self.maker_code



from importlib.metadata import metadata
from typing import Any, cast

from textual import on, work
from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.events import Click
from textual.widgets import Button, Footer, Header, Label
from textual_utils import (
    AppMetadata,
    SettingsScreen,
    mount_about_header_icon,
    translate_about_header_icon,
    translate_bindings,
)
from tilsit_config import load_config, save_settings
from tilsit_i18n import tr

from .app_config import Settings, app_config
from .bindings import NEW_GAME_BINDINGS, GlOBAL_BINDINGS
from .constants import (
    CONFIG_FILE,
    LOCALE_DIR,
)
from .game import Game
from .widgets.board import Board
from .widgets.new_game import NewGameScreen
from .widgets.panel import Panel


class MastermindApp(App[None]):
    CSS_PATH = "styles.tcss"
    ENABLE_COMMAND_PALETTE = False
    BINDINGS = GlOBAL_BINDINGS

    def __init__(self) -> None:
        super().__init__()

        config_dict, settings = cast(
            tuple[dict[str, Any], Settings],
            load_config(config_file=str(CONFIG_FILE), settings_cls=Settings),
        )

        app_config.init(
            ui=config_dict["ui"],
            settings=settings,
            variations=config_dict["variations"],
        )

        pkg_name = cast(str, __package__)
        pkg_metadata = metadata(pkg_name)

        self.app_metadata = AppMetadata(
            name="Mastermind",
            version=pkg_metadata["Version"],
            icon=app_config.ui["icon"],
            description="Break the hidden code",
            author=pkg_metadata["Author"],
            email=pkg_metadata["Author-email"].split("<")[1][:-1],
        )

        tr.localedir = LOCALE_DIR
        tr.language = app_config.settings.language.current_value

        self.panel: Panel
        self.board: Board
        self.game: Game

    def compose(self) -> ComposeResult:
        yield Header()
        yield Horizontal(id="body")
        yield Footer()

    async def on_mount(self) -> None:
        await mount_about_header_icon(
            current_app=self,
            app_metadata=self.app_metadata,
        )

        self.title = self.app_metadata.name

        self.translate()

        self.create_new_game()

    def translate(self) -> None:
        translate_about_header_icon(app=self)
        translate_bindings(screen=self, bindings=GlOBAL_BINDINGS)

    def create_new_game(self) -> None:
        if hasattr(self, "game"):
            self.panel.remove()
            self.board.remove()

        self.panel = Panel()
        self.board = Board()
        body: Horizontal = self.query_one("#body", Horizontal)
        body.mount(self.panel)
        body.mount(self.board)

        self.set_focus(self.panel)

        self.game = Game()

    @on(Button.Pressed, ".code_peg")
    def on_code_peg_pressed(self, event: Button.Pressed):
        active_symbol: int = self.panel.active_symbol
        if active_symbol != 0:
            event.button.label = app_config.ui["code_symbols"][active_symbol - 1]
        else:
            event.button.label = app_config.ui["code_blank_symbol"]

    @on(Click, ".check")
    def on_check_click(self) -> None:
        breaker_code: list[int] = []
        for code_peg in self.board.current_row.code_pegs:
            symbol_str = cast(str, code_peg.label)

            symbol: int
            if symbol_str == app_config.ui["code_blank_symbol"]:
                symbol = 0
            else:
                symbol = app_config.ui["code_symbols"].index(symbol_str) + 1

            breaker_code.append(symbol)

        num_red_pegs: int
        num_white_pegs: int
        num_red_pegs, num_white_pegs = self.game.check_code(breaker_code)

        self.board.current_row.query_one("#check").remove()

        self.board.current_row.mount(
            Label(
                "".join(
                    [
                        (app_config.ui["feedback_symbols"][0] + " ") * num_red_pegs,
                        (app_config.ui["feedback_symbols"][1] + " ") * num_white_pegs,
                        (app_config.ui["feedback_blank_symbol"] + " ")
                        * (self.game.num_pegs - num_red_pegs - num_white_pegs),
                    ]
                ),
                classes="feedback_pegs",
            )
        )

        self.board.current_row.disabled = True

        if num_red_pegs == self.game.num_pegs:
            self.notify(tr("Congratulations!"))
        else:
            if self.board.current_row_number < self.game.num_rows:
                self.board.add_row()
            else:
                maker_code: list[int] = self.game.get_maker_code()
                maker_code_str: str = ""
                for symbol in maker_code:
                    if symbol == 0:
                        maker_code_str += app_config.ui["code_blank_symbol"] + " "
                    else:
                        maker_code_str += (
                            app_config.ui["code_symbols"][symbol - 1] + " "
                        )

                self.notify(
                    f"{tr('Better luck next time')}\n{tr('Code')}: {maker_code_str}",
                    timeout=60,
                )

    @work
    async def action_new_game(self) -> None:
        new_game_screen = NewGameScreen()
        translate_bindings(screen=new_game_screen, bindings=NEW_GAME_BINDINGS)

        if await self.push_screen_wait(new_game_screen):
            self.create_new_game()

            if any(
                [
                    app_config.settings.variation.changed,
                    app_config.settings.duplicate_symbols.changed,
                    app_config.settings.blank_symbol.changed,
                ]
            ):
                save_settings(str(CONFIG_FILE), app_config.settings)

    @work
    async def action_settings(self) -> None:
        if await self.push_screen_wait(
            SettingsScreen(
                dialog_title="Settings",
                dialog_subtitle=self.app_metadata.name,
                settings=[
                    app_config.settings.language,
                ],
            )
        ):
            if app_config.settings.language.changed:
                tr.language = app_config.settings.language.current_value
                self.translate()

                save_settings(str(CONFIG_FILE), app_config.settings)



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Checkbox, Footer, Header, Label, RadioButton, RadioSet
from tilsit_i18n import tr

from ..app_config import app_config
from ..bindings import NEW_GAME_BINDINGS

BORDER_WIDTH = 1
RADIO_SET_LEFT_PADDING_WIDTH = 5
RADIO_SET_RIGHT_PADDING_WIDTH = 2


class NewGameScreen(ModalScreen[bool]):
    BINDINGS = NEW_GAME_BINDINGS

    def compose(self) -> ComposeResult:
        yield Header(icon=app_config.ui["new_game_icon"])

        self.variation_radio_buttons = {
            name: RadioButton(
                label=tr(
                    f"{name} ({variation['num_rows']} rows, {variation['num_pegs']} pegs, {variation['num_symbols']} symbols)"
                )
            )
            for name, variation in app_config.variations.items()
        }

        self.variation_radio_buttons[
            app_config.settings.variation.current_value
        ].value = True

        blank_symbol_str = tr("Blank symbol")
        duplicate_symbols_str = tr("Duplicate symbols")

        labels = [rb.label for rb in self.variation_radio_buttons.values()] + [
            blank_symbol_str,
            duplicate_symbols_str,
        ]

        yield Label(tr("Variation") + ":", classes="margin-bottom-1")
        self.variation_radio_set = RadioSet()
        with self.variation_radio_set:
            for rb in self.variation_radio_buttons.values():
                yield rb

        yield Label(tr("Additional options") + ":", classes="margin-bottom-1")
        self.blank_symbol_cb = Checkbox(
            blank_symbol_str,
            value=app_config.settings.blank_symbol.current_value,
            classes="margin-bottom-1",
        )
        self.duplicate_symbols_cb = Checkbox(
            duplicate_symbols_str,
            value=app_config.settings.duplicate_symbols.current_value,
            classes="margin-bottom-1",
        )

        yield self.blank_symbol_cb
        yield self.duplicate_symbols_cb

        self.variation_radio_set.styles.width = self.blank_symbol_cb.styles.width = (
            self.duplicate_symbols_cb.styles.width
        ) = (
            len(max(labels, key=len))
            + BORDER_WIDTH
            + RADIO_SET_LEFT_PADDING_WIDTH
            + RADIO_SET_RIGHT_PADDING_WIDTH
            + BORDER_WIDTH
        )

        yield Footer()

    def on_mount(self) -> None:
        self.sub_title = tr("New game")

    def action_escape(self) -> None:
        self.dismiss(False)

    def action_next(self) -> None:
        app_config.settings.variation.old_value = (
            app_config.settings.variation.current_value
        )
        app_config.settings.variation.current_value = list(
            self.variation_radio_buttons.keys()
        )[self.variation_radio_set.pressed_index]

        app_config.settings.blank_symbol.old_value = (
            app_config.settings.blank_symbol.current_value
        )
        app_config.settings.blank_symbol.current_value = self.blank_symbol_cb.value

        app_config.settings.duplicate_symbols.old_value = (
            app_config.settings.duplicate_symbols.current_value
        )
        app_config.settings.duplicate_symbols.current_value = (
            self.duplicate_symbols_cb.value
        )

        self.dismiss(True)



from dataclasses import dataclass
from typing import TypedDict

from tilsit_config.settings import SettingBoolean, SettingOptions


class Ui(TypedDict):
    icon: str
    new_game_icon: str
    code_blank_symbol: str
    code_symbols: list[str]
    feedback_blank_symbol: str
    feedback_symbols: list[str]
    check_default_text: str
    check_hover_text: str


@dataclass
class Settings:
    language: SettingOptions
    variation: SettingOptions
    blank_symbol: SettingBoolean
    duplicate_symbols: SettingBoolean


class Variation(TypedDict):
    num_rows: int
    num_pegs: int
    num_symbols: int


class AppConfig:
    ui: Ui
    settings: Settings
    variations: dict[str, Variation]

    def init(self, ui: Ui, settings: Settings, variations: dict[str, Variation]):
        self.ui = ui
        self.settings = settings
        self.variations = variations

    @property
    def variation(self) -> Variation:
        return self.variations[self.settings.variation.current_value]


app_config = AppConfig()



from typing import Final

from textual.binding import Binding

GlOBAL_BINDINGS: Final[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
    Binding(
        key="ctrl+q",
        action="quit",
        description="Quit",
        key_display="Ctrl+Q",
        show=True,
    ),
    Binding(
        key="ctrl+c",
        action="nothing",
        description="",
    ),
    Binding(
        key="f2",
        action="new_game",
        description="New game",
        key_display="F2",
    ),
    Binding(
        key="f3",
        action="settings",
        description="Settings",
        key_display="F3",
    ),
]

NEW_GAME_BINDINGS: Final[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
    Binding(
        key="escape",
        action="escape",
        description="Cancel",
        key_display="Esc",
        show=True,
    ),
    Binding(
        key="greater_than_sign",
        action="next",
        description="Next",
        key_display=">",
        show=True,
    ),
]



from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.message import Message
from textual.widgets import Button

from ..app_config import app_config


class SymbolButton(Button):
    class Toggled(Message):
        def __init__(self, sender: "SymbolButton") -> None:
            super().__init__()
            self.sender = sender

    def on_click(self):
        self.post_message(self.Toggled(sender=self))


class Panel(VerticalScroll):
    def __init__(self) -> None:
        super().__init__()
        self.active_color = 0

    def compose(self) -> ComposeResult:
        variation = app_config.variation

        self.symbol_buttons: list[SymbolButton] = [
            SymbolButton(app_config.ui["code_blank_symbol"], classes="active")
        ]

        self.symbol_buttons.extend(
            [
                SymbolButton(symbol)
                for symbol in app_config.ui["code_symbols"][: variation["num_symbols"]]
            ]
        )

        for symbol_button in self.symbol_buttons:
            yield symbol_button

    def on_symbol_button_toggled(self, message: SymbolButton.Toggled):
        for symbol_button in self.symbol_buttons:
            symbol_button.remove_class("active")

        message.sender.add_class("active")

        self.active_symbol = self.symbol_buttons.index(message.sender)
